{
****************************************************************
*  Name    : SUART.BAS                                         *
*  Author  : David John Barker                                 *
*  Notice  : Copyright (c) 2006 Mecanique                      *
*          : All Rights Reserved                               *
*  Date    : 17/03/2020                                        *
*  Version : 1.5 17/03/2020                                    *
*          : add #option SUART_MODE and SUART_INIT_TX          *
*          : (see notes below)                                 *
*          : 1.4 19/08/2019                                    *
*          : add PORT/LAT/TRIS offset for devices with         *
*          :   different reg maps (K40, K42, Q10, etc)         *
*          :   now works with all but J94/J99 PORTK/PORTL      *
*          : add TX pin init to idle state                     *
*  Notes   : 1.3 28/05/2006 Changed "W" to "WREG" in ReadByte  *
*          : 1.2 Replaced 'rcall' with 'call'                  *
*          : 1.1 Release version                               *
****************************************************************
}
Module UART

Include "system.bas"

// default Mode setting
// valid settings are 'umXXXX' values from the UART mode consts listed below
// set '#option SUART_MODE = 1' to emulate a normal hw uart (ie idle high)
#if isoption(SUART_MODE) and not(SUART_MODE in (1, 2, 4, 4+1, 4+2))
  #error "Invalid SUART_MODE setting"
#endif
#option SUART_MODE = 2     // default = umInverted
const SUART_DEFAULT_MODE = SUART_MODE

// this option controls TX pin intialization in SetTX()
//   true = pin is initialized to idle state
//   false = leave pin uninitialized (v1.3 default)
// note: for this to work properly you must either
//   a) change default '#option SUART_MODE', or
//   b) call SetMode() before calling SetTX()
// if this option is false then the TX pin state is left undefined
// after the call to SetTX()... LAT registers are not initialized at PON/RESET
#if isoption(SUART_INIT_TX) and not(SUART_INIT_TX in (true, false))
  #error "Invalid SUART_INIT_TX setting"
#endif
#option SUART_INIT_TX = true

Const
   CycleCost = 20,                          // TX and RX cycle cost
   FOSC = _clock * 1000000,                 // clock (MHz)
   tcy300 = (2 * FOSC) / (4 * 300) / 2,     // 300 baud Tcy
   tcy600 = (2 * FOSC) / (4 * 600) / 2,     // 600 baud Tcy
   tcy1200 = (2 * FOSC) / (4 * 1200) / 2,   // 1200 baud Tcy
   tcy2400 = (2 * FOSC) / (4 * 2400) / 2,   // 2400 baud Tcy
   tcy4800 = (2 * FOSC) / (4 * 4800) / 2,   // 4800 baud Tcy
   tcy9600 = (2 * FOSC) / (4 * 9600) / 2,   // 9600 baud Tcy
   tcy19200 = (2 * FOSC) / (4 * 19200) / 2, // 19200 baud Tcy
   tcy38400 = (2 * FOSC) / (4 * 38400) / 2, // 38400 baud Tcy - pushing the limits, high xtals only!
   tcy57600 = (2 * FOSC) / (4 * 57600) / 2  // 57600 baud Tcy - pushing the limits, high xtals only!


// calculate baud rate constants - these figures are one
// BitTime (BT) in microseconds (us). For example, 1 / 9600 baud
// gives 104.16 us. Taking into account the TX and RX cycle costs
// yields a figure of 100 us
Public Const
   sbr300 As Word = (tcy300 - CycleCost) / (_clock / 4),
   sbr600 As Word = (tcy600 - CycleCost) / (_clock / 4),
   sbr1200 As Word = (tcy1200 - CycleCost) / (_clock / 4),
   sbr2400 As Word = (tcy2400 - CycleCost) / (_clock / 4),
   sbr4800 As Word = (tcy4800 - CycleCost) / (_clock / 4),
   sbr9600 As Word = (tcy9600 - CycleCost) / (_clock / 4),
   sbr19200 As Word = (tcy19200 - CycleCost) / (_clock / 4),
   sbr38400 As Word = (tcy38400 - CycleCost) / (_clock / 4),
   sbr57600 As Word = (tcy57600 - CycleCost) / (_clock / 4)  

// public UART modes...
Public Const
   umTrue = 1,                          // normal UART mode
   umInverted = 2,
   umOpen = 4,
   umOpenTrue = umOpen Or umTrue,
   umOpenInverted = umOpen Or umInverted 

// a pin structure...
Structure TPin
   PortAddr As Word                   // PORT addr (input), LAT addr (output)
   TrisAddr As Word                   // TRIS addr
   Pin      As Byte
   PinMask  As Byte
End Structure 

// private variables and aliases... 
Dim 
   FTX As TPin,                       // TX pin
   FRX As TPin,                       // RX pin
   FBaudrate As Word,                 // Baudrate (1 BT)
   FHalfBaudrate As Word,             // Half Baudrate (1/2 BT)
   FMode As Byte,                     // Mode (true, inverted, open-drain)
   FPacing As Word,                   // write byte pacing (us)
   FTerminator As Char,               // read string terminator character
   FIsInverted As FMode.Booleans(1)   // inverted mode flag

// public variables...
Public Dim
   ReadTerminator As FTerminator,     // read string terminator
   Pacing As FPacing                  // write pacing (us) 
{
****************************************************************************
* Name    : LAT_OFFSET/TRIS_OFFSET                                         *
* Purpose : these macros load 'wresult' with the LAT and TRIS register     *
*         : offsets from PORT. they assume wresult is a 16-bit word        *
*         : located in the access bank (PROD is a good choice)             *
****************************************************************************
}
macro LAT_OFFSET(wresult)
    asm-
        movlw low(LATA-PORTA)
        movwf wresult
        movlw high(LATA-PORTA)
        movwf wresult+1
    end asm
end macro

macro TRIS_OFFSET(wresult)
    asm-
        movlw low(TRISA-PORTA)
        movwf wresult
        movlw high(TRISA-PORTA)
        movwf wresult+1
    end asm
end macro
{
****************************************************************************
* Name    : SetTX                                                          *
* Purpose : Set TX port and pin number                                     *
****************************************************************************
}
Public Sub SetTX(ByRef pPin As Bit)
   dim reg_offset as PROD      // 16-bit reg (must be in access bank)

   FTX.PortAddr = addressof(pPin)               // PORT addr
   TRIS_OFFSET(reg_offset)                      // get TRIS offset
   FTX.TrisAddr = FTX.PortAddr + reg_offset     // TRIS addr
   LAT_OFFSET(reg_offset)                       // get LAT offset
   FTX.PortAddr = FTX.PortAddr + reg_offset     // change PortAddr to LAT addr

   FTX.Pin = BitOf(pPin)               // get pin number
   FTX.PinMask = Not FTX.Pin           // create pin number mask

 #if (SUART_INIT_TX)
   // set TX pin idle state... (added in v1.4)
   FSR0 = FTX.PortAddr
   If (FIsInverted) Then
      INDF0 = INDF0 And FTX.PinMask   
   Else
      INDF0 = INDF0 Or FTX.Pin
   EndIf
   FSR0 = FTX.TrisAddr
   If ((FMode And umOpen) = 0) Then
      INDF0 = INDF0 And FTX.PinMask     // set to output
   Else
      INDF0 = INDF0 Or FTX.Pin          // set to input
   EndIf
 #endif     //_SUART_INIT_TX
End Sub
{
****************************************************************************
* Name    : SetRX                                                          *
* Purpose : Set RX port and pin number                                     *
****************************************************************************
}
Public Sub SetRX(ByRef pPin As Bit)
   dim reg_offset as PROD      // 16-bit reg (must be in access bank)
   
   FRX.PortAddr = addressof(pPin)               // point address to port
   TRIS_OFFSET(reg_offset)                      // get TRIS offset
   FRX.TrisAddr = FRX.PortAddr + reg_offset     // TRIS addr

   FRX.Pin = BitOf(pPin)               // get pin number
   FRX.PinMask = Not FRX.Pin           // create pin number mask

   FSR0  = FRX.TrisAddr                // point FSR0 to TRIS
   INDF0 = INDF0 Or FRX.Pin            // set to input
End Sub
{
****************************************************************************
* Name    : SetBaudrate                                                    *
* Purpose : Sets the software UART baudrate                                *
*         : Pass a 'sbr' constant, as defined above. For example, sbr9600  *
****************************************************************************
}
Public Sub SetBaudrate(pBaudrate As Word)
   FBaudrate = pBaudrate
   FHalfBaudrate = FBaudrate / 2
End Sub
{
****************************************************************************
* Name    : SetMode                                                        *
* Purpose : Set the UART mode, either TRUE or INVERTED                     *
*         : Use a 'um' constant, as defined above (For example 'umTrue')   *       
*         : Module default is umInverted                                   *
****************************************************************************
}
Public Sub SetMode(pMode As Byte)
   FMode = pMode
End Sub
{
****************************************************************************
* Name    : OneBT (PRIVATE)                                                *
* Purpose : One bit time                                                   *
****************************************************************************
}
NoInline Sub OneBT()
   DelayUS(FBaudrate)
End Sub  
{
****************************************************************************
* Name    : HalfBT (PRIVATE)                                               *
* Purpose : One half bit time                                              *
****************************************************************************
}
NoInline Sub HalfBT()
   DelayUS(FHalfBaudrate)
End Sub 
{
****************************************************************************
* Name    : ReadByte                                                       *
* Purpose : Read a byte from the software UART                             *
****************************************************************************
}
Public Function Access ReadByte() As Byte
   Dim Index, Pin, Mode As Byte
   
   // get port, pin and mode...
   Pin = FRX.Pin
   FSR0 = FRX.PortAddr      // PORT address
   Mode = FMode
   Index = 8  
  
ASM-
   ; look For start Bit...
   movf    INDF0, W         ; <Read Bit>
   btfsc   Mode, 1                 
   comf    WREG             ; <- changed from "W"
   andwf   Pin, W                      
   btfss   STATUS, Z               
   bra     $ - 10                     
End ASM
   HalfBT()                   
ASM-
   movf    INDF0, W         ; <Read Bit>
   btfsc   Mode, 1          
   comf    WREG             ; <- changed from "W"
   andwf   Pin, W           
   btfss   STATUS, Z        
   bra     $ - 24 

   ; Read data bits...
   call    OneBT            ; 2 + BT + 2  | 20
   movf    INDF0, W         ; 1           | ** <Read Pin>
   bcf     STATUS, C        ; 1           | 1
   andwf   Pin, W           ; 1           | 2
   btfss   STATUS, Z        ; 1 (2)       | 3 (4)
   bsf     STATUS, C        ; 1           | 4
   rrcf    Result, F        ; 1           | 5
   bra     $ + 2            ; 2           | 7
   bra     $ + 2            ; 2           | 9
   bra     $ + 2            ; 2           | 11
   bra     $ + 2            ; 2           | 13
   decfsz  Index, F         ; 1 (2)       | 14 (15)
   bra     $ - 26           ; 2           | 16
End ASM
   OneBT()
   If FIsInverted Then
      Result = Not Result
   EndIf  
   ClrWDT   
End Function

{
****************************************************************************
* Name    : WriteByte                                                      *
* Purpose : Write a byte to the software UART                              *
****************************************************************************
}
Public Sub Access WriteByte(pValue As Byte)  
   Dim Index, Mode, ModeOpen As Byte
   Dim Pin, PinMask As Byte
   Dim MakeHigh As Pin, MakeLow As PinMask
   Dim MakeInput As Pin, MakeOutput As PinMask
   Dim SaveTRIS As ModeOpen
   Dim DATAPort As INDF0
   Dim TRISPort As INDF1
   
   // set mode...
   Mode = FMode
   ModeOpen = 0
   If FIsInverted Then
      ModeOpen = FTX.Pin
      pValue = Not pValue
   EndIf  
   Pin = FTX.Pin              // pin mask
   PinMask = FTX.PinMask      // inverse mask
   FSR0 = FTX.PortAddr        // LAT address
   FSR1 = FTX.TrisAddr        // TRIS address
   
   Index = 9                  // 1 start, 8 data bits
ASM-
   ; start Bit...
   bsf    STATUS, C       ; start Bit is High
   btfsc  Mode, 0         ; is it true mode
   bcf    STATUS, C       ; yes, start Bit is Low
   bra    $ + 4           ; GoTo Write Bit

   ; main loop...
   rrcf   pValue, F       ; 1           | 8

   ; set direction...  
   movf   TRISPort, W     ; 1           | 9
   andwf  MakeOutput, W   ; 1           | 10   
   btfsc  STATUS, C       ; 1 (2)       | 11 (12)
   iorwf  MakeInput, W    ; 1           | 12 
   xorwf  ModeOpen, W     ; 1           | 13
   btfss  Mode, 2         ; 1 (2)       | 14 (15)
   andwf  MakeOutput, W   ; 1           | 15
   movwf  TRISPort        ; 1           | 16 <set tris>

   ; set Bit value...
   movf   DataPort, W     ; 1           | 17
   andwf  MakeLow, W      ; 1           | 18 
   btfsc  STATUS, C       ; 1 (2)       | 19 (20)
   iorwf  MakeHigh, W     ; 1           | 20 
   movwf  DataPort        ; 1           | ** <Write Pin>
   call   OneBT           ; 2 + BT + 2  | 4
   decfsz Index, F        ; 1 (2)       | 5 (6)
   bra    $ - 34          ; 2           | 7

   ; stop Bit, set direction...
   movf   TRISPort, W     ; 1           | 8
   iorwf  MakeInput, W    ; 1           | 9
   btfss  Mode, 2         ; 1 (2)       | 10 (11)
   andwf  MakeOutput, W   ; 1           | 11
   movwf  SaveTRIS        ; 1           | 12 

   ; stop Bit, data...
   movf   INDF0, W        ; 1           | 13
   andwf  MakeLow, W      ; 1           | 14
   btfsc  Mode, 0         ; 1 (2)       | 15 (16)
   iorwf  MakeHigh, W     ; 1           | 16
   bra    $ + 2           ; 2           | 18
   movff  SaveTRIS, INDF1 ; 2           | 20 <set tris>
   movwf  INDF0           ; 1           | ** <Write Pin>
End ASM

   // return from ASM and make direct call - also ensures
   // that subroutine is linked into program to allow above
   // call from within the ASM block...
   OneBT()
   If (FPacing > 0) Then
      DelayUS(FPacing)
   EndIf 
   ClrWDT  
End Sub
{
****************************************************************************
* Name    : ReadBoolean                                                    *
* Purpose : Read a boolean from the software UART                          *
****************************************************************************
}
Public Function ReadBoolean() As Boolean
   Result = Boolean(ReadByte)
End Function 
{
****************************************************************************
* Name    : ReadWord                                                       *
* Purpose : Read a word from the software UART                             *
****************************************************************************
}
Public Function ReadWord() As Word
   Result.Bytes(0) = ReadByte
   Result.Bytes(1) = ReadByte
End Function  
{
****************************************************************************
* Name    : ReadLongWord                                                   *
* Purpose : Read a long word from the software UART                        *
****************************************************************************
}
Public Function ReadLongWord() As LongWord
   Result.Bytes(0) = ReadByte
   Result.Bytes(1) = ReadByte
   Result.Bytes(2) = ReadByte
   Result.Bytes(3) = ReadByte
End Function  
{
****************************************************************************
* Name    : ReadFloat                                                      *
* Purpose : Read a float from the software UART                            *
****************************************************************************
}
Public Function ReadFloat() As Float
   Result.Bytes(0) = ReadByte
   Result.Bytes(1) = ReadByte
   Result.Bytes(2) = ReadByte
   Result.Bytes(3) = ReadByte
End Function  
{
****************************************************************************
* Name    : WriteBoolean                                                   *
* Purpose : Write a boolean value to the software UART                     *
****************************************************************************
}
Public Sub WriteBoolean(pValue As Boolean)
   WriteByte(Byte(pValue))
End Sub
{
****************************************************************************
* Name    : WriteWord                                                      *
* Purpose : Write a word value to the software UART                        *
****************************************************************************
}
Public Sub WriteWord(pValue As Word)
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
End Sub
{
****************************************************************************
* Name    : WriteLongWord                                                  *
* Purpose : Write a word value to the software UART                        *
****************************************************************************
}
Public Sub WriteLongWord(pValue As LongWord) 
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
   WriteByte(pValue.Bytes(2))
   WriteByte(pValue.Bytes(3))
End Sub
{
****************************************************************************
* Name    : WriteFloat                                                     *
* Purpose : Write a floating point number to the software UART             *
****************************************************************************
}
Public Sub WriteFloat(pValue As Float) 
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
   WriteByte(pValue.Bytes(2))
   WriteByte(pValue.Bytes(3))
End Sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a boolean from the software UART                          *
****************************************************************************
}
Sub ReadItem(ByRef pValue As Boolean)
   pValue = Boolean(ReadByte)
End Sub 
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a byte from the software UART                             *
****************************************************************************
}
Sub ReadItem(ByRef pValue As Byte)
   pValue = ReadByte
End Sub  
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a shortint from the software UART                         *
****************************************************************************
}
Sub ReadItem(ByRef pValue As ShortInt)
   pValue = ReadByte
End Sub  
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a word from the software UART                             *
****************************************************************************
}
Sub ReadItem(ByRef pValue As Word)
   pValue = ReadWord
End Sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read an integer from the software UART                         *
****************************************************************************
}
Sub ReadItem(ByRef pValue As Integer)
   pValue = ReadWord
End Sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a long word from the software UART                        *
****************************************************************************
}
Sub ReadItem(ByRef pValue As LongWord)
   pValue = ReadLongWord
End Sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a long integer from the software UART                     *
****************************************************************************
}
Sub ReadItem(ByRef pValue As LongInt)
   pValue = ReadLongWord
End Sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a floating point number from the software UART            *
****************************************************************************
}
Sub ReadItem(ByRef pValue As Float) 
   pValue = ReadFloat
End Sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a string from the software UART. Use ReadTerminator       *
*         : to specify the input string terminator character.              *
****************************************************************************
}
Sub ReadItem(ByRef pText As String)
   Dim Data, Index As Byte
   
   Index = 0
   Data = ReadByte
   While Data <> Byte(FTerminator)
      pText(Index) = Data
      Data = ReadByte
      Inc(Index)
   Wend
   pText(Index) = null
End Sub
{
****************************************************************************
* Name    : Read (COMPOUND)                                                *
* Purpose : Read an item from the software UART                            *
****************************************************************************
}
Public Compound Sub Read(ReadItem)
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a boolean value to the software UART                     *
****************************************************************************
}
Sub WriteItem(pValue As Boolean)
   WriteByte(Byte(pValue))
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a byte value to the software UART                        *
****************************************************************************
}
Sub WriteItem(pValue As WREG)
   WriteByte(pValue)
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a short int value to the software UART                   *
****************************************************************************
}
Sub WriteItem(pValue As ShortInt)
   WriteByte(pValue)
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a word value to the software UART                        *
****************************************************************************
}
Sub WriteItem(pValue As Word) 
   WriteWord(pValue)
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write an integer value to the software UART                    *
****************************************************************************
}
Sub WriteItem(pValue As Integer) 
   WriteWord(pValue)
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a long word to the software UART                         *
****************************************************************************
}
Sub WriteItem(pValue As LongWord) 
   WriteLongWord(pValue)
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a long integer to the software UART                      *
****************************************************************************
}
Sub WriteItem(pValue As LongInt) 
   WriteLongWord(pValue)
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a floating point number to the software UART             *
****************************************************************************
}
Sub WriteItem(pValue As Float) 
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
   WriteByte(pValue.Bytes(2))
   WriteByte(pValue.Bytes(3))
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a string value to the software UART                      *
****************************************************************************
}
Sub WriteItem(pText As String)
   Dim Index, Data As Byte  
   Index = 0
   Data = pText(0)
   While Data <> 0
      WriteByte(Data)
      Inc(Index)
      Data = pText(index)
   Wend
End Sub
{
****************************************************************************
* Name    : Write (COMPOUND)                                               *
* Purpose : Write an item to the software UART                             *
****************************************************************************
}
Public Compound Sub Write(WriteItem)

// initialise module...
FTerminator = #13
FPacing = 0
FMode = SUART_DEFAULT_MODE

End Module

